///////////////////////////////////////////////////////////////////////////////
// PureCrypt v1.0a                                                           //
///////////////////////////////////////////////////////////////////////////////
// Blowfish cipher routines.                                                 //
///////////////////////////////////////////////////////////////////////////////

// The usual suspects
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
// Endian specific macros - converting to a big endian machine would need to be
// done here
#include "System.h"
// Bits and pieces relevant for the functions below
#include "Blowfish.h"
// Mostly just needed for the HeapAlloc() function
#include <windows.h>

// The PB memory heap handle
extern int *PB_MemoryBase;

// PureCrypt API //////////////////////////////////////////////////////////////

// These are probably leaking memory like a sieve due to the fact that they return
// a new block of allocated memory each time they are called.  Might have to somehow
// check to see if heap has been allocated and free it before creating a new one.
// Maybe handle it a bit like strok(). ie. 

// Blowfish encode 8 bytes
PBCHAR PB_BlowfishEncode(uchar8 *password, ulong32 length, uchar8 *buffer, int rounds)
{
  BLOWFISH key;
  char *result = NULL;

  // Setup the number of rounds
  switch (rounds) {
    case 0:
      rounds = 0;
    case 1:
      rounds = 16;
    default:
      rounds = 0;
  }

  // Have we got input?
  if (buffer && *buffer && password && *password && length > 0) {
    if (blowfish_setup(password, length, rounds, &key) != TRUE) {
      return result;
    }
    // Allocate memory within the PB heap
    result = HeapAlloc(PB_MemoryBase, HEAP_ZERO_MEMORY, 8);
    if (blowfish_ecb_encrypt(buffer, (unsigned char *)result, &key) != TRUE) {
      // Free the allocated memory
      HeapFree(PB_MemoryBase, 0, result);
      result = NULL;
      return result;
    }
  }

  return result;
}

// Blowfish decode 8 bytes
PBCHAR PB_BlowfishDecode(uchar8 *password, ulong32 length, uchar8 *buffer, int rounds)
{
  BLOWFISH key;
  char *result = NULL;

  // Setup the number of rounds
  switch (rounds) {
    case 0:
      rounds = 0;
    case 1:
      rounds = 16;
    default:
      rounds = 0;
  }

  // Have we got input?
  if (buffer && *buffer && password && *password && length > 0) {
    if (blowfish_setup(password, length, rounds, &key) != TRUE) {
      return result;
    }
    // Allocate memory within the PB heap
    result = HeapAlloc(PB_MemoryBase, HEAP_ZERO_MEMORY, 8);
    if (blowfish_ecb_decrypt(buffer, (unsigned char *)result, &key) != TRUE) {
      // Free the allocated memory
      HeapFree(PB_MemoryBase, 0, result);
      result = NULL;
      return result;
    }
  }

  return result;
}

// Blowfish ///////////////////////////////////////////////////////////////////

int blowfish_setup(uchar8 *key, int keylen, int num_rounds, BLOWFISH *skey)
{
  ulong32 x, y, z, A;
  uchar8 B[8];

  if (key == NULL || skey == NULL)
    return FALSE;

  /* check key length */
  if (keylen < 8 || keylen > 56) {
    return FALSE;
  }

  /* check rounds */
  if (num_rounds != 0 && num_rounds != 16) {
    return FALSE;
  }   

  /* load in key bytes (Supplied by David Hopwood) */
  for (x = y = 0; x < 18; x++) {
    A = 0;
    for (z = 0; z < 4; z++) {
      A = (A << 8) | ((ulong32)key[y++] & 255);
      if (y == (ulong32)keylen) { 
        y = 0; 
      }
    }
    skey->K[x] = ORIG_P[x] ^ A;
  }

  /* copy sboxes */
  for (x = 0; x < 4; x++) {
    for (y = 0; y < 256; y++) {
      skey->S[x][y] = ORIG_S[x][y];
    }
  }

  /* encrypt K array */
  for (x = 0; x < 8; x++) {
    B[x] = 0;
  }
   
  for (x = 0; x < 18; x += 2) {
    /* encrypt it */
    blowfish_ecb_encrypt(B, B, skey);
    /* copy it */
    LOAD32H(skey->K[x], &B[0]);
    LOAD32H(skey->K[x+1], &B[4]);
  }

  /* encrypt S array */
  for (x = 0; x < 4; x++) {
    for (y = 0; y < 256; y += 2) {
      /* encrypt it */
      blowfish_ecb_encrypt(B, B, skey);
      /* copy it */
      LOAD32H(skey->S[x][y], &B[0]);
      LOAD32H(skey->S[x][y+1], &B[4]);
    }
  }

  return TRUE;
}

#ifdef S1
#undef S1
#endif
#define S1 skey->S[0]
#ifdef S2
#undef S2
#endif
#define S2 skey->S[1]
#ifdef S3
#undef S3
#endif
#define S3 skey->S[2]
#ifdef S4
#undef S4
#endif
#define S4 skey->S[3]

#ifdef F
#undef F
#endif
#define F(x) ((S1[byte(x,3)] + S2[byte(x,2)]) ^ S3[byte(x,1)]) + S4[byte(x,0)]

int blowfish_ecb_encrypt(uchar8 *pt, uchar8 *ct, BLOWFISH *skey)
{
  ulong32 L, R;
  int r;

  if (pt == NULL || ct == NULL || skey == NULL)
    return FALSE;

  /* load it */
  LOAD32H(L, &pt[0]);
  LOAD32H(R, &pt[4]);

  /* do 16 rounds */
  for (r = 0; r < 16; ) {
    L ^= skey->K[r++];  R ^= F(L);
    R ^= skey->K[r++];  L ^= F(R);
    L ^= skey->K[r++];  R ^= F(L);
    R ^= skey->K[r++];  L ^= F(R);
  }

  /* last keying */
  R ^= skey->K[17];
  L ^= skey->K[16];

  /* store */
  STORE32H(R, &ct[0]);
  STORE32H(L, &ct[4]);

  return TRUE;
}

int blowfish_ecb_decrypt(uchar8 *ct, uchar8 *pt, BLOWFISH *skey)
{
  ulong32 L, R;
  int r;

  if (pt == NULL || ct == NULL || skey == NULL)
    return FALSE;

  /* load it */
  LOAD32H(R, &ct[0]);
  LOAD32H(L, &ct[4]);

  /* undo last keying */
  R ^= skey->K[17];
  L ^= skey->K[16];

  /* do 16 rounds */
  for (r = 15; r > 0; ) {
    L ^= F(R); R ^= skey->K[r--];
    R ^= F(L); L ^= skey->K[r--];
    L ^= F(R); R ^= skey->K[r--];
    R ^= F(L); L ^= skey->K[r--];
  }

  /* store */
  STORE32H(L, &pt[0]);
  STORE32H(R, &pt[4]);

  return TRUE;
}
