///////////////////////////////////////////////////////////////////////////////
// PureCrypt v1.0a                                                           //
///////////////////////////////////////////////////////////////////////////////
// SHA-512 hash routines.                                                    //
///////////////////////////////////////////////////////////////////////////////

// The usual suspects
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
// Endian specific macros - converting to a big endian machine would need to be
// done here
#include "System.h"
// Bits and pieces relevant for the functions below
#include "SHA-512.h"
// Mostly just needed for the HeapAlloc() function
#include <windows.h>

// The PB memory heap handle
extern int *PB_MemoryBase;

// PureCrypt API //////////////////////////////////////////////////////////////

// Returns a 64 byte buffer with the result
PBCHAR PB_SHA512Hash(uchar8 *buffer, ulong32 length)
{
  SHA512 md;
  char *result = NULL;

  // Have we got input?
  if (buffer && *buffer) {
    if (sha512_init(&md) != TRUE) {
      return result;
    }
    if (sha512_process(&md, buffer, length) != TRUE) {
      return result;
    }
    // Allocate some memory within the PB heap
    result = HeapAlloc(PB_MemoryBase, HEAP_ZERO_MEMORY, 64);
    if (sha512_done(&md, (unsigned char *)result) != TRUE) {
      // Free the allocated memory
      HeapFree(PB_MemoryBase, 0, result);
      result = NULL;
      return result;
    }
  }

  return result;
}

// SHA-512 ////////////////////////////////////////////////////////////////////

/* Various logical functions */
#ifdef Ch
#undef Ch
#endif
#define Ch(x,y,z)       (z ^ (x & (y ^ z)))
#ifdef Maj
#undef Maj
#endif
#define Maj(x,y,z)      (((x | y) & z) | (x & y)) 
#ifdef S
#undef S
#endif
#define S(x, n)         ROR64c(x, n)
#ifdef R
#undef R
#endif
#define R(x, n)         (((x)&CONST64(0xFFFFFFFFFFFFFFFF))>>((ulong64)n))
#ifdef Sigma0
#undef Sigma0
#endif
#define Sigma0(x)       (S(x, 28) ^ S(x, 34) ^ S(x, 39))
#ifdef Sigma1
#undef Sigma1
#endif
#define Sigma1(x)       (S(x, 14) ^ S(x, 18) ^ S(x, 41))
#ifdef Gamma0
#undef Gamma0
#endif
#define Gamma0(x)       (S(x, 1) ^ S(x, 8) ^ R(x, 7))
#ifdef Gamma1
#undef Gamma1
#endif
#define Gamma1(x)       (S(x, 19) ^ S(x, 61) ^ R(x, 6))

/* compress 1024-bits */
int sha512_compress(SHA512 *md, uchar8 *buf)
{
  ulong64 S[8], W[80], t0, t1;
  int i;

  /* copy state into S */
  for (i = 0; i < 8; i++) {
    S[i] = md->state[i];
  }

  /* copy the state into 1024-bits into W[0..15] */
  for (i = 0; i < 16; i++) {
    LOAD64H(W[i], buf + (8*i));
  }

  /* fill W[16..79] */
  for (i = 16; i < 80; i++) {
    W[i] = Gamma1(W[i - 2]) + W[i - 7] + Gamma0(W[i - 15]) + W[i - 16];
  }        

  /* Compress */
#ifdef RND
#undef RND
#endif
#define RND(a,b,c,d,e,f,g,h,i)                    \
  t0 = h + Sigma1(e) + Ch(e, f, g) + K[i] + W[i];   \
  t1 = Sigma0(a) + Maj(a, b, c);                  \
  d += t0;                                        \
  h  = t0 + t1;

  for (i = 0; i < 80; i += 8) {
    RND(S[0],S[1],S[2],S[3],S[4],S[5],S[6],S[7],i+0);
    RND(S[7],S[0],S[1],S[2],S[3],S[4],S[5],S[6],i+1);
    RND(S[6],S[7],S[0],S[1],S[2],S[3],S[4],S[5],i+2);
    RND(S[5],S[6],S[7],S[0],S[1],S[2],S[3],S[4],i+3);
    RND(S[4],S[5],S[6],S[7],S[0],S[1],S[2],S[3],i+4);
    RND(S[3],S[4],S[5],S[6],S[7],S[0],S[1],S[2],i+5);
    RND(S[2],S[3],S[4],S[5],S[6],S[7],S[0],S[1],i+6);
    RND(S[1],S[2],S[3],S[4],S[5],S[6],S[7],S[0],i+7);
  }

  /* feedback */
  for (i = 0; i < 8; i++) {
    md->state[i] = md->state[i] + S[i];
  }

  return TRUE;
}

int sha512_init(SHA512 *md)
{
  if (md == NULL)
    return FALSE;

  md->curlen = 0;
  md->length = 0;
  md->state[0] = CONST64(0x6a09e667f3bcc908);
  md->state[1] = CONST64(0xbb67ae8584caa73b);
  md->state[2] = CONST64(0x3c6ef372fe94f82b);
  md->state[3] = CONST64(0xa54ff53a5f1d36f1);
  md->state[4] = CONST64(0x510e527fade682d1);
  md->state[5] = CONST64(0x9b05688c2b3e6c1f);
  md->state[6] = CONST64(0x1f83d9abfb41bd6b);
  md->state[7] = CONST64(0x5be0cd19137e2179);

  return TRUE;
}

int sha512_done(SHA512 *md, uchar8 *out)
{
  int i;

  if (md == NULL || out == NULL)
    return FALSE;

  if (md->curlen >= sizeof(md->buf)) {
    return FALSE;
  }

  /* increase the length of the message */
  md->length += md->curlen * CONST64(8);

  /* append the '1' bit */
  md->buf[md->curlen++] = (uchar8)0x80;

  /* if the length is currently above 112 bytes we append zeros
   * then compress.  Then we can fall back to padding zeros and length
   * encoding like normal.
   */
  if (md->curlen > 112) {
    while (md->curlen < 128) {
      md->buf[md->curlen++] = (uchar8)0;
    }
    sha512_compress(md, md->buf);
    md->curlen = 0;
  }

  /* pad upto 120 bytes of zeroes 
   * note: that from 112 to 120 is the 64 MSB of the length.  We assume that you won't hash
   * > 2^64 bits of data... :-)
   */
  while (md->curlen < 120) {
    md->buf[md->curlen++] = (unsigned char)0;
  }

  /* store length */
  STORE64H(md->length, md->buf+120);
  sha512_compress(md, md->buf);

    /* copy output */
  for (i = 0; i < 8; i++) {
    STORE64H(md->state[i], out+(8*i));
  }

  return TRUE;
}

int sha512_process(SHA512 *md, uchar8 *in, ulong32 inlen)
{
  ulong32 n;
  int blocksize = 128;

  // Check to see we have been passed valid arguments
  if (md == NULL || in == NULL) {
    return FALSE;
  }
  if (md->curlen > sizeof(md->buf)) {
    return FALSE;
  }

  // Process block
  while (inlen > 0) {
    if (md->curlen == 0 && inlen >= blocksize) {
      if ((sha512_compress(md, (unsigned char *)in)) != TRUE) {
        return FALSE;
      }
      md->length += blocksize * 8;
      in += blocksize;
      inlen -= blocksize;
    } else {
      n = MIN(inlen, (blocksize - md->curlen));
      memcpy(md->buf + md->curlen, in, (size_t)n);
      md->curlen += n;
      in += n;
      inlen -= n;
      if (md->curlen == blocksize) {
        if ((sha512_compress(md, md->buf)) != TRUE) {
          return FALSE;
        }
        md->length += 8*blocksize;
        md->curlen = 0;
      }
    }
  }
  return TRUE;
}
